# disjointset.tcl --
#
#  Implementation of a Disjoint Set for Tcl.
#
# 
# Copyright (c) Summer 2008 Alejandro Eduardo Cruz Paz

package require Tcl 8.2
package require struct::set

namespace eval ::struct {}
namespace eval ::struct::disjointset {

  # This is the structure where each disjoint set will be kept. 
  # Each list is a set. 
  variable djsetstructure
  set djsetstructure {}
  
  # Counter for tracking the number of sets actually on the container.
  variable counter 0

  variable commands [list \
        "sets" \
        "cardinality"   \
        "union"   \
        "equals"   \
        "makeset" \
        "findunitereplace"   \
        "find"   \
        "destroy" \
        ]

  namespace export disjointset

}

proc ::struct::disjointset {{name ""}} {
  variable djsetstructure

  # create name if not given
  if {![info exists name]} {
      incr counter
      set name "disjointset${counter}"
  }

  if { ![string equal [info commands ::$name] ""] } {
    error "command \"$name\" already exists, unable to create disjointset"
  }

  # Create the command to manipulate the DisjointSet
  interp alias {} ::$name {} ::struct::disjointset::DisjointSetProc $name
  return $name
}

proc ::struct::disjointset::DisjointSetProc {name {cmd ""} args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 2 } {
    error "wrong # args: should be \"$name option ?arg arg ...?\""
    }

    # Split the args into command and args components
    if { [string equal [info commands ::struct::disjointset::_$cmd] ""] } {
    variable commands
    set optlist [join $commands ", "]
    set optlist [linsert $optlist "end-1" "or"]
    error "bad option \"$cmd\": must be $optlist"
    }
    return [eval [linsert $args 0 ::struct::disjointset::_$cmd $name]]
}

# Creates a new set inside this disjointset structure, verifying the integrity
# of each new insertion for previous existance on the structure
proc ::struct::disjointset::_makeset {name item} {
  if {[_Isunarylist? $name $item]} {
    if {[_find $name $item -index] != -1} {
      return -code error "The item $item is already on this disjoint set"
    }
  } else {
    foreach element $item {
      if {[_find $name $element -index] != -1} {
        return -code error "The item $element in list $item is already\
                            on this disjoint set"
      }
    }
  }

  variable djsetstructure
  variable counter
  incr counter
  set djsetstructure [linsert $djsetstructure 0 $item]
}

proc ::struct::disjointset::_sets {name} {
  variable djsetstructure
  return $djsetstructure
}

# Returns the number of sets of this disjointset
proc ::struct::disjointset::_cardinality {name} {
  variable counter
  return $counter
}

# Joins two sets. If -withincrement option is present, then it will increment the counter of the
# number of sets in the set structure.
#
proc ::struct::disjointset::_union {name set1 set2 {option ""}} {
  set union [::struct::set union $set1 $set2]
  # the union of the same set is the set itself
  if {$option eq {-withincrement}} {
    variable counter
    if {$counter > 1} {
      set counter [expr $counter - 1]
    }
  }
  return $union
}

proc ::struct::disjointset::_equals {name set1 set2} {
  return [::struct::set equal $set1 $set2]
}

proc ::struct::disjointset::_findunitereplace {name item1 item2} {
  variable djsetstructure
  foreach item [concat $item1 $item2] {
    set index [_find $name $item -index]
    lappend setmp [lindex $djsetstructure $index]
    set djsetstructure [lreplace $djsetstructure $index $index]
  }
  lappend djsetstructure [_union $name [lindex $setmp 0] [lindex $setmp 1] {-withincrement}]
}

proc ::struct::disjointset::_find {name item option1} {
  variable djsetstructure
  # search for the item in the disjointset
  set candidate [lsearch $djsetstructure $item] 
  # if it was not found, 
  if {$candidate == -1 && [_Isunarylist? $name $item]} {
    for {set i 0} {$i < [llength $djsetstructure]} {incr i} {
      set set_ [lindex $djsetstructure $i] 
      set candidate [lsearch $set_ $item]
      if {$candidate != -1} {
        set candidate $i
        break
      }
    }
  } 
  switch $option1 {
   {-sets}  { return [lindex $djsetstructure $candidate]}
   {-index} { return $candidate }
   {default} { error "incorrect option, an option needs to be specified, -sets|-index"}
  }
}

proc ::struct::disjointset::_Isunarylist? {name element} {
  if {[llength $element] == 1} {
    return 1
  } else {
    return 0
  }
}

proc ::struct::disjointset::_destroy {name} {
    variable djsetstructure
    variable counter
    set djsetstructure {}
    set counter 0
    interp alias {} ::$name {}
    return
}

namespace eval ::struct {

    namespace import -force disjointset::DisjointSetProc
    namespace export disjointset
}
package provide struct::disjointset 1.0