#ifndef __GNUC__
#error "GCC is needed to compile this file"
#endif

#include <sys/param.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <string.h>

#include "rtld.h"
#include "defs.h"

#include "rtld_machdep.h"

#define PATH_RTLD	"/usr/libexec/ld-elf.so.2"
#define LD_ARY_CACHE	16

/* Types. */
/** Pointer to a function which has no arguments and returns void */
typedef void (*func_ptr_type)();
/** Pointer to a function which takes three arguemnts: const char*, size_t, void*
  * and returns void*.
  * This type of function is used as callback in Tcl_PathEnumerate
  */
typedef void * (*path_enum_proc) (const char *path, size_t len, void *arg);

/**
 * This structure provides a reentrant way to keep a list of objects and
 * check which ones have already been processed in some way.
 */
typedef struct Struct_DoneList {
    const Obj_Entry **objs;		///< Array of object pointers
    unsigned int num_alloc;		///< Allocated size of the array
    unsigned int num_used;		///< Number of array slots used
} DoneList;

/*
 * Function declarations.
 */
static void die(void);
static void Tcl_DigestDynamic(Obj_Entry *);
static Obj_Entry *dlcheck(void *);
static bool donelist_check(DoneList *, const Obj_Entry *);
static void errmsg_restore(char *);
static char *errmsg_save(void);
static char *Tcl_FindLibrary(const char *, const Obj_Entry *);
static Obj_Entry *Tcl_FindObject(const char *);
static const char *gethints(void);
static void init_dag(Obj_Entry *);
static void init_dag1(Obj_Entry *root, Obj_Entry *obj, DoneList *);
static int Tcl_LoadNeededObjects(Obj_Entry *);
static Obj_Entry *Tcl_LoadObjectFromDisk(char *);
static Obj_Entry *Tcl_LoadObjectFromMemory(const void *addr, int nbytes);
static Obj_Entry *Tcl_ObjFromAddr(const void *);
static void objlist_call_fini(Objlist *);
static void objlist_call_init(Objlist *);
static void objlist_clear(Objlist *);
static Objlist_Entry *objlist_find(Objlist *, const Obj_Entry *);
static void objlist_init(Objlist *);
static void objlist_push_head(Objlist *, Obj_Entry *);
static void objlist_push_tail(Objlist *, Obj_Entry *);
static void objlist_remove(Objlist *, Obj_Entry *);
static void *Tcl_PathEnumerate(const char *, path_enum_proc, void *);
static int Tcl_RelocateObjects(Obj_Entry *, bool, Obj_Entry *);
static char *Tcl_SearchLibraryPath(const char *, const char *);
static const Elf_Sym *Tcl_SymlookDefault(const char *, unsigned long hash,
  const Obj_Entry *refobj, const Obj_Entry **defobj_out, bool in_plt);
static const Elf_Sym *Tcl_SymlookList(const char *, unsigned long,
  const Objlist *, const Obj_Entry **, bool in_plt, DoneList *);
static const Elf_Sym *Tcl_SymlookNeeded(const char *, unsigned long,
  const Needed_Entry *, const Obj_Entry **, bool in_plt, DoneList *);
static void Tcl_UnlinkObject(Obj_Entry *);
static void Tcl_UnloadObject(Obj_Entry *);
static void Tcl_UnrefDag(Obj_Entry *);
void Tcl_InitPltGot_i386(Obj_Entry *obj);


/*
 * Data declarations.
 */
static char *error_message;	/* Message for dlerror(), or NULL */
static bool trust;		/* False for setuid and setgid programs */
static const char *ld_bind_now;	/* Environment variable for immediate binding */
static const char *ld_debug;	/* Environment variable for debugging */
static const char *ld_library_path; /* Environment variable for search path */
static char *ld_preload;	/* Environment variable for libraries to
				   load first */
static const char *ld_tracing;	/* Called from ldd(1) to print libs */
				/* Optional function call tracing hook */
static int (*rtld_functrace)(const char *caller_obj,
			     const char *callee_obj,
			     const char *callee_func,
			     void *stack);
static Obj_Entry *rtld_functrace_obj;	/* Object thereof */
static Obj_Entry *obj_list;	/* Head of linked list of shared objects */
static Obj_Entry *obj_tail;	/* Link field of last object in list */
static Obj_Entry **preload_tail;
static Obj_Entry *obj_main;	/* The main program shared object */
static Obj_Entry obj_rtld;	/* The dynamic linker shared object */
static unsigned int obj_count;	/* Number of objects in obj_list */
static int	ld_resident;	/* Non-zero if resident */
static const char *ld_ary[LD_ARY_CACHE];
static int	ld_index;
static Objlist initlist;

static Objlist list_global =	/* Objects dlopened with RTLD_GLOBAL */
  STAILQ_HEAD_INITIALIZER(list_global);
static Objlist list_main =	/* Objects loaded at program startup */
  STAILQ_HEAD_INITIALIZER(list_main);
static Objlist list_fini =	/* Objects needing fini() calls */
  STAILQ_HEAD_INITIALIZER(list_fini);

static Elf_Sym sym_zero;	/* For resolving undefined weak refs. */

#pragma weak _DYNAMIC

/*
 * Global declarations normally provided by crt1.  The dynamic linker is
 * not built with crt1, so we have to provide them ourselves.
 */
//char **environ;

/**
  \return Returns a string which contains the latest error message returned by another function.
*/
const char *
Tcl_Dlerror(void)
{
    char *msg = error_message;
    error_message = NULL;
    return msg;
}


/*
 * Globals to control TLS allocation.
 */
size_t tls_last_offset;		/* Static TLS offset of last module */
size_t tls_last_size;		/* Static TLS size of last module */
size_t tls_static_space;	/* Static TLS space allocated */
int tls_dtv_generation = 1;	/* Used to detect when dtv size changes  */
int tls_max_index = 1;		/* Largest module index allocated */

/*
 * Fill in a DoneList with an allocation large enough to hold all of
 * the currently-loaded objects.  Keep this as a macro since it calls
 * alloca and we want that to occur within the scope of the caller.
 */
#define donelist_init(dlp)					\
    ((dlp)->objs = alloca(obj_count * sizeof (dlp)->objs[0]),	\
    assert((dlp)->objs != NULL),				\
    (dlp)->num_alloc = obj_count,				\
    (dlp)->num_used = 0)

/**
 * Return a dynamically-allocated copy of the current error message, if any.
 * \return newly allocated string that is a copy of the current error message
 */
static char *
errmsg_save(void)
{
    return error_message == NULL ? NULL : xstrdup(error_message);
}

/**
 * Restore the current error message from a copy which was previously saved
 * by errmsg_save(). The copy is freed.
 * \param saved_msg error message that will be resored
 */
static void
errmsg_restore(char *saved_msg)
{
    if (saved_msg == NULL)
	error_message = NULL;
    else {
	printf("%s", saved_msg);
	free(saved_msg);
    }
}

/**
 * Returns only file name from a given path
 * \param name given path
 * \return file name filtred from the path
 */
const char *
basename(const char *name)
{
    const char *p = strrchr(name, '/');
    return p != NULL ? p + 1 : name;
}

/**
 * Function that is invoked if a serious error occued.
 * Execution of program should be terminated
 */
static void
die(void)
{
    const char *msg = Tcl_Dlerror();

    if (msg == NULL)
	msg = "Fatal error";
    errx(1, "%s", msg);
}

/**
 * Process a shared object's DYNAMIC section, and save the important
 * information in its Obj_Entry structure.
 * \param *obj pointer to a shared object whose DYNAMIC section will be processed
 * \callgraph
 */
static void
Tcl_DigestDynamic(Obj_Entry *obj)
{
    const Elf_Dyn *dynp;
    Needed_Entry **needed_tail = &obj->needed;
    const Elf_Dyn *dyn_rpath = NULL;
    int plttype = DT_REL;
	printf("DIGEST_DYNAMIC: BEGIN %u %u\n", obj->dynamic,  obj->dynamic->d_tag);
    for (dynp = obj->dynamic;  dynp->d_tag != DT_NULL;  dynp++) {
	switch (dynp->d_tag) {

	case DT_REL:
	    obj->rel = (const Elf_Rel *) (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_RELSZ:
	    obj->relsize = dynp->d_un.d_val;
	    break;

	case DT_RELENT:
	    assert(dynp->d_un.d_val == sizeof(Elf_Rel));
	    break;

	case DT_JMPREL:
	    obj->pltrel = (const Elf_Rel *)
	      (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_PLTRELSZ:
	    obj->pltrelsize = dynp->d_un.d_val;
	    break;

	case DT_RELA:
	    obj->rela = (const Elf_Rela *) (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_RELASZ:
	    obj->relasize = dynp->d_un.d_val;
	    break;

	case DT_RELAENT:
	    assert(dynp->d_un.d_val == sizeof(Elf_Rela));
	    break;

	case DT_PLTREL:
	    plttype = dynp->d_un.d_val;
	    assert(dynp->d_un.d_val == DT_REL || plttype == DT_RELA);
	    break;

	case DT_SYMTAB:
	    obj->symtab = (const Elf_Sym *)
	      (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_SYMENT:
	    assert(dynp->d_un.d_val == sizeof(Elf_Sym));
	    break;

	case DT_STRTAB:
	    obj->strtab = (const char *) (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_STRSZ:
	    obj->strsize = dynp->d_un.d_val;
	    break;

	case DT_HASH:
	    {
		const Elf_Addr *hashtab = (const Elf_Addr *)
		  (obj->relocbase + dynp->d_un.d_ptr);
		obj->nbuckets = hashtab[0];
		obj->nchains = hashtab[1];
		obj->buckets = hashtab + 2;
		obj->chains = obj->buckets + obj->nbuckets;
	    }
	    break;

	case DT_NEEDED:
	    if (!obj->rtld) {
		Needed_Entry *nep = NEW(Needed_Entry);
		nep->name = dynp->d_un.d_val;
		nep->obj = NULL;
		nep->next = NULL;

		*needed_tail = nep;
		needed_tail = &nep->next;
	    }
	    break;

	case DT_PLTGOT:
	    obj->pltgot = (Elf_Addr *) (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_TEXTREL:
	    obj->textrel = true;
	    break;

	case DT_SYMBOLIC:
	    obj->symbolic = true;
	    break;

	case DT_RPATH:
	case DT_RUNPATH:	/* XXX: process separately */
	    /*
	     * We have to wait until later to process this, because we
	     * might not have gotten the address of the string table yet.
	     */
	    dyn_rpath = dynp;
	    break;

	case DT_SONAME:
	    /* Not used by the dynamic linker. */
	    break;

	case DT_INIT:
	    obj->init = (InitFunc) (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_FINI:
	    obj->fini = (InitFunc) (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_FLAGS:
		if (dynp->d_un.d_val & DF_ORIGIN) {
		    obj->origin_path = xmalloc(PATH_MAX);
		}
		if (dynp->d_un.d_val & DF_SYMBOLIC)
		    obj->symbolic = true;
		if (dynp->d_un.d_val & DF_TEXTREL)
		    obj->textrel = true;
		if (dynp->d_un.d_val & DF_BIND_NOW)
		    obj->bind_now = true;
		if (dynp->d_un.d_val & DF_STATIC_TLS)
		    ;
	    break;

	default:
		printf("Ignoring d_tag %d = %#x\n", dynp->d_tag, dynp->d_tag);
	    break;
	}
    }

    obj->traced = false;

    if (plttype == DT_RELA) {
	obj->pltrela = (const Elf_Rela *) obj->pltrel;
	obj->pltrel = NULL;
	obj->pltrelasize = obj->pltrelsize;
	obj->pltrelsize = 0;
    }

    if (dyn_rpath != NULL)
	obj->rpath = obj->strtab + dyn_rpath->d_un.d_val;
	printf("DIGEST_DYNAMIC: END\n");
}

static Obj_Entry *
dlcheck(void *handle)
{
    Obj_Entry *obj;
	printf("OBJ_LIST: %u\n", obj_list->next);
    for (obj = obj_list;  obj != NULL;  obj = obj->next)
	if (obj == (Obj_Entry *) handle)
	    break;
	printf("DLCHECK %u %u %u %u\n", obj_list, obj, obj->refcount, obj->dl_refcount);
    if (obj == NULL || obj->refcount == 0 || obj->dl_refcount == 0) {
	printf("Invalid shared object handle %p", handle);
	return NULL;
    }
    return obj;
}

/**
 * If the given object is already in the donelist, return true.  Otherwise
 * add the object to the list and return false.
 * \param dlp list of shared objects
 * \param obj a shared object that we want to find in the list
 * \return true if the object was aleady in the dlp list, false otherwise
 */
static bool
donelist_check(DoneList *dlp, const Obj_Entry *obj)
{
    unsigned int i;

    for (i = 0;  i < dlp->num_used;  i++)
	if (dlp->objs[i] == obj)
	    return true;
    /*
     * Our donelist allocation should always be sufficient.  But if
     * our threads locking isn't working properly, more shared objects
     * could have been loaded since we allocated the list.  That should
     * never happen, but we'll handle it properly just in case it does.
     */
    if (dlp->num_used < dlp->num_alloc)
	dlp->objs[dlp->num_used++] = obj;
    return false;
}

/**
 * Hash function for symbol table lookup.  Don't even think about changing
 * this.  It is specified by the System V ABI.
 * \param  name name of a symbol
 * \return returns a value of a hash function for its argument
 */
unsigned long
elf_hash(const char *name)
{
    const unsigned char *p = (const unsigned char *) name;
    unsigned long h = 0;
    unsigned long g;

    while (*p != '\0') {
	h = (h << 4) + *p++;
	if ((g = h & 0xf0000000) != 0)
	    h ^= g >> 24;
	h &= ~g;
    }
    return h;
}

/**
 * Find the library with the given name, and return its full pathname.
 * The returned string is dynamically allocated.  Generates an error
 * message and returns NULL if the library cannot be found.
 *
 * If the second argument is non-NULL, then it refers to an already-
 * loaded shared object, whose library search path will be searched.
 *
 * The search order is:
 *   LD_LIBRARY_PATH
 *   rpath in the referencing file
 *   ldconfig hints
 *   /lib
 * \param name library name
 * \param refobj referencing shared object
 * \return full path of the library
 */
static char *
Tcl_FindLibrary(const char *name, const Obj_Entry *refobj)
{
    char *pathname;

    if (strchr(name, '/') != NULL) {	/* Hard coded pathname */
	if (name[0] != '/' && !trust) {
	    printf("Absolute pathname required for shared object \"%s\"",
	      name);
	    return NULL;
	}
	return xstrdup(name);
    }

    printf(" Searching for \"%s\"", name);

    if ((pathname = Tcl_SearchLibraryPath(name, ld_library_path)) != NULL ||
      (refobj != NULL &&
      (pathname = Tcl_SearchLibraryPath(name, refobj->rpath)) != NULL) ||
      (pathname = Tcl_SearchLibraryPath(name, gethints())) != NULL ||
      (pathname = Tcl_SearchLibraryPath(name, STANDARD_LIBRARY_PATH)) != NULL)
	return pathname;

    if(refobj != NULL && refobj->path != NULL) {
	printf("Shared object \"%s\" not found, required by \"%s\"",
	  name, basename(refobj->path));
    } else {
	printf("Shared object \"%s\" not found", name);
    }
    return NULL;
}

/**
 * Given a symbol number in a referencing object, find the corresponding
 * definition of the symbol.  Returns a pointer to the symbol, or NULL if
 * no definition was found.  Returns a pointer to the Obj_Entry of the
 * defining object via the reference parameter DEFOBJ_OUT.
 * \param symnum symbol number
 * \param refobj referencing shared object
 * \param defobj_out
 * \param in_plt
 * \return pointer to the symbol
 */
const Elf_Sym *
Tcl_FindSymdef(unsigned long symnum, const Obj_Entry *refobj,
    const Obj_Entry **defobj_out, bool in_plt, SymCache *cache)
{
    const Elf_Sym *ref;
    const Elf_Sym *def;
    const Obj_Entry *defobj;
    const char *name;
    unsigned long hash;

    /*
     * If we have already found this symbol, get the information from
     * the cache.
     */
	printf("ELF_SYM: begin\n");
    if (symnum >= refobj->nchains)
	return NULL;	/* Bad object */
    if (cache != NULL && cache[symnum].sym != NULL) {
	*defobj_out = cache[symnum].obj;
	return cache[symnum].sym;
    }

    ref = refobj->symtab + symnum;
    name = refobj->strtab + ref->st_name;
    defobj = NULL;
	printf("ELF_SYM: 01 name %s\n", name);
    /*
     * We don't have to do a full scale lookup if the symbol is local.
     * We know it will bind to the instance in this load module; to
     * which we already have a pointer (ie ref). By not doing a lookup,
     * we not only improve performance, but it also avoids unresolvable
     * symbols when local symbols are not in the hash table.
     *
     * This might occur for TLS module relocations, which simply use
     * symbol 0.
     */
    if (ELF_ST_BIND(ref->st_info) != STB_LOCAL) {
	if (ELF_ST_TYPE(ref->st_info) == STT_SECTION) {
	    printf("%s: Bogus symbol table entry %lu", refobj->path,
		symnum);
	}
	hash = elf_hash(name);
	def = Tcl_SymlookDefault(name, hash, refobj, &defobj, in_plt);
	printf("ELF_SYM: 1.5 def = %u\n", def);
    } else {
	def = ref;
	defobj = refobj;
    }
	printf("ELF_SYM: 02\n");
    /*
     * If we found no definition and the reference is weak, treat the
     * symbol as having the value zero.
     */
    if (def == NULL && ELF_ST_BIND(ref->st_info) == STB_WEAK) {
	printf("ELF_SYM: 02.5\n");
	def = &sym_zero;
	defobj = obj_main;
    }
	printf("ELF_SYM: 03\n");
    if (def != NULL) {
	*defobj_out = defobj;
	/* Record the information in the cache to avoid subsequent lookups. */
	if (cache != NULL) {
	    cache[symnum].sym = def;
	    cache[symnum].obj = defobj;
	}
	printf("ELF_SYM: 04\n");
    } else
	printf("%s: Undefined symbol \"%s\"", refobj->path, name);
    return def;
}

/**
 * Return the search path from the ldconfig hints file, reading it if
 * necessary.  Returns NULL if there are problems with the hints file,
 * or if the search path there is empty.
 * \return a newly allocated search path found from the ldconfig hints file
 */
static const char *
gethints(void)
{
    static char *hints;

    if (hints == NULL) {
	int fd;
	struct elfhints_hdr hdr;
	char *p;

	/* Keep from trying again in case the hints file is bad. */
	hints = "";

	if ((fd = open(_PATH_ELF_HINTS, O_RDONLY)) == -1)
	    return NULL;
	if (read(fd, &hdr, sizeof hdr) != sizeof hdr ||
	  hdr.magic != ELFHINTS_MAGIC ||
	  hdr.version != 1) {
	    close(fd);
	    return NULL;
	}
	p = xmalloc(hdr.dirlistlen + 1);
	if (lseek(fd, hdr.strtab + hdr.dirlist, SEEK_SET) == -1 ||
	  read(fd, p, hdr.dirlistlen + 1) != hdr.dirlistlen + 1) {
	    free(p);
	    close(fd);
	    return NULL;
	}
	hints = p;
	close(fd);
    }
    return hints[0] != '\0' ? hints : NULL;
}

/**
 * This function creates a Directed Aciclic Graph whose root will be a given
 * shared objects and nodes will be other shared objects that depend on each other.
 * For example: if one object is a child of the other, then this object is needed by its parent 
 * \param root a shared object that will be the root of the new DAG
 */
static void
init_dag(Obj_Entry *root)
{
    DoneList donelist;

    donelist_init(&donelist);
    init_dag1(root, root, &donelist);
}

/**
 * Function that is responsible for actual creation of DAG
 */
static void
init_dag1(Obj_Entry *root, Obj_Entry *obj, DoneList *dlp)
{
    const Needed_Entry *needed;

    if (donelist_check(dlp, obj))
	return;
    objlist_push_tail(&obj->dldags, root);
    objlist_push_tail(&root->dagmembers, obj);
    for (needed = obj->needed;  needed != NULL;  needed = needed->next)
	if (needed->obj != NULL)
	    init_dag1(root, needed->obj, dlp);
}

/**
 * Given a shared object, traverse its list of needed objects, and load
 * each of them.  Returns 0 on success.  Generates an error message and
 * returns -1 on failure.
 * \param first pointer to the first loaded shared objects
 * \return 0 on success, -1 on failure
 */
static int
Tcl_LoadNeededObjects(Obj_Entry *first)
{
    Obj_Entry *obj;
	printf("Tcl_LoadNeededObjects: begin\n");
    for (obj = first;  obj != NULL;  obj = obj->next) {
	Needed_Entry *needed;
	printf("Tcl_LoadNeededObjects: loop %s\n", obj->path);
	for (needed = obj->needed;  needed != NULL;  needed = needed->next) {
	    const char *name = obj->strtab + needed->name;
		printf("Tcl_LoadNeededObjects: name %s\n", name);
	    char *path = Tcl_FindLibrary(name, obj);
	    needed->obj = NULL;
	    if (path == NULL && !ld_tracing)
		return -1;

	    if (path) {
		needed->obj = Tcl_LoadObjectFromDisk(path);
		if (needed->obj == NULL && !ld_tracing)
		    return -1;		/* XXX - cleanup */
	    }
	}
    }
	printf("Tcl_LoadNeededObjects: end\n");
    return 0;
}

#define RTLD_FUNCTRACE "_rtld_functrace"

/**
 * Returns a pointer to the Obj_Entry for the object with the given path.
 * Returns NULL if no matching object was found.
 * \param path path to a shared on disk
 * \return pointer to the shared object structure in memory
 */
static Obj_Entry *
Tcl_FindObject(const char *path)
{
    Obj_Entry *obj;

    for (obj = obj_list->next;  obj != NULL;  obj = obj->next) {
	if (obj->path && strcmp(obj->path, path) == 0)
	    return(obj);
    }
    return(NULL);
}


/**
 * Load a shared object into memory, if it is not already loaded.  The
 * argument must be a string allocated on the heap.  This function assumes
 * responsibility for freeing it when necessary.
 * \param path path to the shared object file on disk
 * \return returns a pointer to the Obj_Entry for the object.  Returns NULL
 * on failure.
 */
static Obj_Entry *
Tcl_LoadObjectFromDisk(char *path)
{
    Obj_Entry *obj = NULL;
    int fd = -1;
    struct stat sb;
	printf("LOAD_OBJECT: BEGIN %s\n", path);
	obj = Tcl_FindObject(path);
	printf("LOAD_OBJECT: 0\n");
    if (obj != NULL) {
	obj->refcount++;
	free(path);
	return(obj);
    }
	printf("LOAD_OBJECT: 1\n");
    if ((fd = open(path, O_RDONLY)) == -1) {
	free(path);
	return (NULL);
    }
	printf("LOAD_OBJECT: 2 FD %d\n", fd);
	printf("LOAD_OBJECT: 3\n");
    printf("loading \"%s\"", path);
    obj = Tcl_MapObjectFromFile(fd, path, &sb);
    close(fd);
    if (obj == NULL) {
	free(path);
        return NULL;
    }
	printf("LOAD OBJECT 4\n");
    obj->path = path;
    Tcl_DigestDynamic(obj);
	printf("LOAD OBJECT 5\n");
    printf("LOAD OBJECT 5.a OBJ_TAIL %u\n", obj_tail);
    obj_tail->next = obj;
    obj->prev = obj_tail;
    obj_tail = obj;
    //TODO obj_tail = &obj->next;
    obj_count++;
    //linkmap_add(obj);	/* for GDB & dlinfo() */

    printf("  %p .. %p: %s", obj->mapbase, obj->mapbase + obj->mapsize - 1,
	obj->path);
    if (obj->textrel)
        printf("  WARNING: %s has impure text", obj->path);

    obj->refcount++;
    return obj;
}

/**
 * Load a shared object into memory, if it is not already loaded.  The
 * argument must be a pointer to the shared object that is already in memory.
 * \param addr pointer to the shared object in memory
 * \param nbytes size of the shared object in bytes
 * \return returns a pointer to the Obj_Entry for the object.  Returns NULL
 * on failure.
 */
static Obj_Entry *
Tcl_LoadObjectFromMemory(const void *addr, int nbytes)
{
    Obj_Entry *obj = NULL;
    int fd = -1;
    struct stat sb;
	printf("LOAD_OBJECT: BEGIN\n");
	printf("LOAD_OBJECT FROM MEMORY: 0\n");
    obj = Tcl_MapObjectFromMemory(addr, nbytes);
    if (obj == NULL) {
        return NULL;
    }
	printf("LOAD OBJECT 4\n");
    Tcl_DigestDynamic(obj);
	printf("LOAD OBJECT 5 %u\n", obj_tail->next);
    obj_tail->next = obj;
    obj->prev = obj_tail;
    obj_tail = obj;
    obj_count++;
    obj->refcount++;
    return obj;
}

/**
 * Find a Obj_Entry struct for a shared object that was loaded at a given address
 * \param addr address of a shared object
 * \return Obj_Entry struct for the shared object or NULL if no object was found
 */
static Obj_Entry *
Tcl_ObjFromAddr(const void *addr)
{
    Obj_Entry *obj;
	printf("Tcl_ObjFromAddr: begin\n");
    for (obj = obj_list;  obj != NULL;  obj = obj->next) {
	printf("Tcl_ObjFromAddr: loop\n");
	if (addr < (void *) obj->mapbase)
	    continue;
	if (addr < (void *) (obj->mapbase + obj->mapsize))
	    return obj;
    }
    return NULL;
}

/**
 * Call the finalization functions for each of the objects in "list"
 * which are unreferenced.  All of the objects are expected to have
 * non-NULL fini functions.
 * \param list list of shared objects
 */
static void
objlist_call_fini(Objlist *list)
{
    Objlist_Entry *elm;
    char *saved_msg;

    /*
     * Preserve the current error message since a fini function might
     * call into the dynamic linker and overwrite it.
     */
    saved_msg = errmsg_save();
    STAILQ_FOREACH(elm, list, link) {
	if (elm->obj->refcount == 0) {
	    printf("calling fini function for %s", elm->obj->path);
	    (*elm->obj->fini)();
	}
    }
    errmsg_restore(saved_msg);
}

/**
 * Call the initialization functions for each of the objects in
 * "list".  All of the objects are expected to have non-NULL init
 * functions.
 * \param list list of shared objects
 */
static void
objlist_call_init(Objlist *list)
{
    Objlist_Entry *elm;
    char *saved_msg;

    /*
     * Preserve the current error message since an init function might
     * call into the dynamic linker and overwrite it.
     */
    saved_msg = errmsg_save();
    STAILQ_FOREACH(elm, list, link) {
	printf("calling init function for %s", elm->obj->path);
	(*elm->obj->init)();
    }
    errmsg_restore(saved_msg);
}

/**
 * Remove all objects from the a given "list". Memory will be freed.
 * \param list list of shared objects
 */
static void
objlist_clear(Objlist *list)
{
    Objlist_Entry *elm;

    while (!STAILQ_EMPTY(list)) {
	elm = STAILQ_FIRST(list);
	STAILQ_REMOVE_HEAD(list, link);
	free(elm);
    }
}

/**
 * Find a shared object in a given "list"
 * \param list list of shared objects
 * \param obj object that we want to find in the list
 * \return Objlist_Entry struct for obj if the object was found or NULL otherwise
 */
static Objlist_Entry *
objlist_find(Objlist *list, const Obj_Entry *obj)
{
    Objlist_Entry *elm;

    STAILQ_FOREACH(elm, list, link)
	if (elm->obj == obj)
	    return elm;
    return NULL;
}

/**
 * Initialize a "list" of shared objects
 * \param list list of shared objects
 */
static void
objlist_init(Objlist *list)
{
    STAILQ_INIT(list);
}

/**
 * Add a shared object at the begining of a given list
 * \param list list of shared objects
 * \param obj a shared object that will be added to the list
 */
static void
objlist_push_head(Objlist *list, Obj_Entry *obj)
{
    Objlist_Entry *elm;

    elm = NEW(Objlist_Entry);
    elm->obj = obj;
    STAILQ_INSERT_HEAD(list, elm, link);
}

/**
 * Add a shared object at the end of a given list
 * \param list list of shared objects
 * \param obj a shared object that will be added to the list
 */
static void
objlist_push_tail(Objlist *list, Obj_Entry *obj)
{
    Objlist_Entry *elm;

    elm = NEW(Objlist_Entry);
    elm->obj = obj;
    STAILQ_INSERT_TAIL(list, elm, link);
}

/**
 * Remove a shared object from a given list
 * \param list list of shared objects
 * \param obj a shared object that will be removed from the list
 */
static void
objlist_remove(Objlist *list, Obj_Entry *obj)
{
    Objlist_Entry *elm;

    if ((elm = objlist_find(list, obj)) != NULL) {
	STAILQ_REMOVE(list, elm, Struct_Objlist_Entry, link);
	free(elm);
    }
}

/**
 * Relocate newly-loaded shared objects.  The argument is a pointer to
 * the Obj_Entry for the first such object.  All objects from the first
 * to the end of the list of objects are relocated.  Returns 0 on success,
 * or -1 on failure.
 * \param first
 * \param bind_now
 * \param rtldobj
 */
static int
Tcl_RelocateObjects(Obj_Entry *first, bool bind_now, Obj_Entry *rtldobj)
{
    Obj_Entry *obj;
	printf("Tcl_RelocateObjects: begin\n");
    for (obj = first;  obj != NULL;  obj = obj->next) {
	printf("Tcl_RelocateObjects: loop\n");
	if (obj != rtldobj)
	    printf("relocating \"%s\"\n", obj->path);
	
	if (obj->nbuckets == 0 || obj->nchains == 0 || obj->buckets == NULL ||
	    obj->symtab == NULL || obj->strtab == NULL) {
	    printf("%s: Shared object has no run-time symbol table",
	      obj->path);
	    return -1;
	}
	printf("Tcl_RelocateObjects: 02\n");
	if (obj->textrel) {
	    /* There are relocations to the write-protected text segment. */
	    if (mprotect(obj->mapbase, obj->textsize,
	      PROT_READ|PROT_WRITE|PROT_EXEC) == -1) {
		printf("%s: Cannot write-enable text segment: %s",
		  obj->path, strerror(errno));
		return -1;
	    }
	}
	printf("Tcl_RelocateObjects: 03\n");
	/* Process the non-PLT relocations. */
	if (Tcl_RelocNonPlt_i386(obj, rtldobj))
		return -1;
	printf("Tcl_RelocateObjects: 04\n");
	/*
	 * Reprotect the text segment.  Make sure it is included in the
	 * core dump since we modified it.  This unfortunately causes the
	 * entire text segment to core-out but we don't have much of a
	 * choice.  We could try to only reenable core dumps on pages
	 * in which relocations occured but that is likely most of the text
	 * pages anyway, and even that would not work because the rest of
	 * the text pages would wind up as a read-only OBJT_DEFAULT object
	 * (created due to our modifications) backed by the original OBJT_VNODE
	 * object, and the ELF coredump code is currently only able to dump
	 * vnode records for pure vnode-backed mappings, not vnode backings
	 * to memory objects.
	 */
	if (obj->textrel) {
	    printf("Tcl_RelocateObjects: 05\n");
	    madvise(obj->mapbase, obj->textsize, MADV_CORE);
	    printf("Tcl_RelocateObjects: 06\n");
	    if (mprotect(obj->mapbase, obj->textsize,
	      PROT_READ|PROT_EXEC) == -1) {
		printf("%s: Cannot write-protect text segment: %s",
		  obj->path, strerror(errno));
		return -1;
	    }
	}
	printf("Tcl_RelocateObjects: 07\n");
	/* Process the PLT relocations. */
	if (Tcl_RelocPlt_i386(obj) == -1)
	    return -1;
	printf("Tcl_RelocateObjects: 08\n");
	/* Relocate the jump slots if we are doing immediate binding. */
	if (obj->bind_now || bind_now)
	    if (Tcl_RelocJmpslots_i386(obj) == -1)
		return -1;


	/*
	 * Set up the magic number and version in the Obj_Entry.  These
	 * were checked in the crt1.o from the original ElfKit, so we
	 * set them for backward compatibility.
	 */
	obj->magic = RTLD_MAGIC;
	obj->version = RTLD_VERSION;
	printf("Tcl_RelocateObjects: 09\n");
	/* Set the special PLT or GOT entries. */
	Tcl_InitPltGot_i386(obj);
	printf("Tcl_RelocateObjects: 10\n");
    }
    printf("Tcl_RelocateObjects: END\n");
    return 0;
}

/*
 * Cleanup procedure.  It will be called (by the atexit mechanism) just
 * before the process exits.
 */

static void *
Tcl_PathEnumerate(const char *path, path_enum_proc callback, void *arg)
{
    if (path == NULL)
	return (NULL);

    path += strspn(path, ":;");
    while (*path != '\0') {
	size_t len;
	char  *res;

	len = strcspn(path, ":;");
	res = callback(path, len, arg);

	if (res != NULL)
	    return (res);

	path += len;
	path += strspn(path, ":;");
    }

    return (NULL);
}

struct try_library_args {
    const char	*name;
    size_t	 namelen;
    char	*buffer;
    size_t	 buflen;
};

static void *
Tcl_TryLibraryPath(const char *dir, size_t dirlen, void *param)
{
    struct try_library_args *arg;

    arg = param;
    if (*dir == '/' || trust) {
	char *pathname;

	if (dirlen + 1 + arg->namelen + 1 > arg->buflen)
		return (NULL);

	pathname = arg->buffer;
	strncpy(pathname, dir, dirlen);
	pathname[dirlen] = '/';
	strcpy(pathname + dirlen + 1, arg->name);

	printf("  Trying \"%s\"", pathname);
	if (access(pathname, F_OK) == 0) {		/* We found it */
	    pathname = xmalloc(dirlen + 1 + arg->namelen + 1);
	    strcpy(pathname, arg->buffer);
	    return (pathname);
	}
    }
    return (NULL);
}

static char *
Tcl_SearchLibraryPath(const char *name, const char *path)
{
    char *p;
    struct try_library_args arg;

    if (path == NULL)
	return NULL;

    arg.name = name;
    arg.namelen = strlen(name);
    arg.buffer = xmalloc(PATH_MAX);
    arg.buflen = PATH_MAX;

    p = Tcl_PathEnumerate(path, Tcl_TryLibraryPath, &arg);

    free(arg.buffer);

    return (p);
}

/**
 * Unload a shared object that was loaded by Tcl_Dlopen. If there was only one
 * reference to this object, then it is removed and memory is freed.
 * \param handle a handler to a shared object that was loaded by Tcl_Dlopen
 * \return 0 on success, -1 on failure
 */
int
Tcl_Dlclose(void *handle)
{
    Obj_Entry *root;

    //wlock_acquire();
     root = dlcheck(handle);
     printf("ELF_DLCOSE: 01\n");
     if (root == NULL) {
	//wlock_release();
	return -1;
     }
 printf("ELF_DLCOSE: 02\n");
     /* Unreference the object and its dependencies. */
     root->dl_refcount--;
     Tcl_UnrefDag(root);
 printf("ELF_DLCOSE: 03\n");
     if (root->refcount == 0) {
	/*
	 * The object is no longer referenced, so we must unload it.
	 * First, call the fini functions with no locks held.
	 */
	//wlock_release();
	//objlist_call_fini(&list_fini);
	//wlock_acquire();
	printf("ELF_DLCOSE: 04\n");
	/* Finish cleaning up the newly-unreferenced objects. */
	Tcl_UnloadObject(root);
	printf("ELF_DLCOSE: 05\n");
    }
   //wlock_release();
    return 0;
}

void *
Tcl_Dlopen(const void *addr, int nbytes, int mode)
{
    Obj_Entry *obj;
    Objlist initlist;
    int result;

    obj = NULL;
     obj = Tcl_LoadObjectFromMemory(addr, nbytes);

     if (obj) {
	obj->dl_refcount++;
	    result = Tcl_LoadNeededObjects(obj);
		printf("Tcl_Dlopen: 13 RESULT = %d\n", result);
	    if (result == -1 ||
	      (init_dag(obj), Tcl_RelocateObjects(obj, mode == RTLD_NOW,
	       &obj_rtld)) == -1) {
		if (obj->refcount == 0)
		    Tcl_UnloadObject(obj);
		obj = NULL;
	    }
    }
    return obj;
}

/**
 * Find address of a symbol given by name in a shared object that is given by its handle that
 * was returned by Tcl_Dlopen when this object was loaded.
 * \param handle address of a shared object that was given byl Tcl_Dlopen
 * \param name name of a symbol that we want to find
 * \return address of the symbol of NULL if the symbol was not found
 */
void *
Tcl_Dlsym(void *handle, const char *name)
{
    const Obj_Entry *obj;
    unsigned long hash;
    const Elf_Sym *def;
    const Obj_Entry *defobj;

     hash = elf_hash(name);
     def = NULL;
     defobj = NULL;

     if (handle == NULL) {
	void *retaddr;

	retaddr = __builtin_return_address(0);	/* __GNUC__ only */
	printf("Tcl_Dlsym: 02 retaddr %u\n", retaddr);
	if ((obj = Tcl_ObjFromAddr(retaddr)) == NULL) {
	    printf("Cannot determine caller's shared object");
	    //TODO rlock_release();
	    return NULL;
	}
     } else {
	DoneList donelist;
	printf("Tcl_Dlsym: 02b\n");
	if (((Obj_Entry *)handle)->magic != RTLD_MAGIC) {
		return NULL;
	}
	if ((obj = dlcheck(handle)) == NULL) {
	    return NULL;
	}
	donelist_init(&donelist);
	if (obj->mainprog) {
	    /* Search main program and all libraries loaded by it. */
	    def = Tcl_SymlookList(name, hash, &list_main, &defobj, true,
	      &donelist);
	} else {
	    Needed_Entry fake;
	    /* Search the given object and its needed objects. */
	    fake.next = NULL;
	    fake.obj = (Obj_Entry *)obj;
	    fake.name = 0;
	    def = Tcl_SymlookNeeded(name, hash, &fake, &defobj, true,
	      &donelist);
	}
    }

    if (def != NULL) {
	return defobj->relocbase + def->st_value;
    }
    return NULL;
}

/*
 * Given a symbol name in a referencing object, find the corresponding
 * definition of the symbol.  Returns a pointer to the symbol, or NULL if
 * no definition was found.  Returns a pointer to the Obj_Entry of the
 * defining object via the reference parameter DEFOBJ_OUT.
 */
static const Elf_Sym *
Tcl_SymlookDefault(const char *name, unsigned long hash,
    const Obj_Entry *refobj, const Obj_Entry **defobj_out, bool in_plt)
{
    DoneList donelist;
    const Elf_Sym *def;
    const Elf_Sym *symp;
    const Obj_Entry *obj;
    const Obj_Entry *defobj;
    const Objlist_Entry *elm;
    def = NULL;
    defobj = NULL;
    donelist_init(&donelist);

    /* Look first in the referencing object if linked symbolically. */
    if (refobj->symbolic && !donelist_check(&donelist, refobj)) {
	symp = Tcl_SymlookObj(name, hash, refobj, in_plt);
	if (symp != NULL) {
	    def = symp;
	    defobj = refobj;
	}
    }

    /* Search all objects loaded at program start up. */
    if (def == NULL || ELF_ST_BIND(def->st_info) == STB_WEAK) {
	symp = Tcl_SymlookList(name, hash, &list_main, &obj, in_plt, &donelist);
	if (symp != NULL &&
	  (def == NULL || ELF_ST_BIND(symp->st_info) != STB_WEAK)) {
	    def = symp;
	    defobj = obj;
	}
    }

    /* Search all DAGs whose roots are RTLD_GLOBAL objects. */
    STAILQ_FOREACH(elm, &list_global, link) {
       if (def != NULL && ELF_ST_BIND(def->st_info) != STB_WEAK)
           break;
       symp = Tcl_SymlookList(name, hash, &elm->obj->dagmembers, &obj, in_plt,
         &donelist);
	if (symp != NULL &&
	  (def == NULL || ELF_ST_BIND(symp->st_info) != STB_WEAK)) {
	    def = symp;
	    defobj = obj;
	}
    }
    /* Search all dlopened DAGs containing the referencing object. */
    STAILQ_FOREACH(elm, &refobj->dldags, link) {

	if (def != NULL && ELF_ST_BIND(def->st_info) != STB_WEAK)
	    break;
	symp = Tcl_SymlookList(name, hash, &elm->obj->dagmembers, &obj, in_plt,
	  &donelist);

	if (symp != NULL &&
	  (def == NULL || ELF_ST_BIND(symp->st_info) != STB_WEAK)) {
	    def = symp;
	    defobj = obj;
	}
    }

    /*
     * Search the dynamic linker itself, and possibly resolve the
     * symbol from there.  This is how the application links to
     * dynamic linker services such as dlopen.  Only the values listed
     * in the "exports" array can be resolved from the dynamic linker.
     */
    if (def == NULL || ELF_ST_BIND(def->st_info) == STB_WEAK) {
	symp = Tcl_SymlookObj(name, hash, &obj_rtld, in_plt);
	if (symp != NULL) {
	    def = symp;
	    defobj = &obj_rtld;
	}
    }

    if (def != NULL)
	*defobj_out = defobj;
    return def;
}

static const Elf_Sym *
Tcl_SymlookList(const char *name, unsigned long hash, const Objlist *objlist,
  const Obj_Entry **defobj_out, bool in_plt, DoneList *dlp)
{
    const Elf_Sym *symp;
    const Elf_Sym *def;
    const Obj_Entry *defobj;
    const Objlist_Entry *elm;

    def = NULL;
    defobj = NULL;
    STAILQ_FOREACH(elm, objlist, link) {
	if (donelist_check(dlp, elm->obj))
	    continue;
	if ((symp = Tcl_SymlookObj(name, hash, elm->obj, in_plt)) != NULL) {
	    if (def == NULL || ELF_ST_BIND(symp->st_info) != STB_WEAK) {
		def = symp;
		defobj = elm->obj;
		if (ELF_ST_BIND(def->st_info) != STB_WEAK)
		    break;
	    }
	}
    }
    if (def != NULL)
	*defobj_out = defobj;
    return def;
}

/*
 * Search the symbol table of a shared object and all objects needed
 * by it for a symbol of the given name.  Search order is
 * breadth-first.  Returns a pointer to the symbol, or NULL if no
 * definition was found.
 */
static const Elf_Sym *
Tcl_SymlookNeeded(const char *name, unsigned long hash, const Needed_Entry *needed,
  const Obj_Entry **defobj_out, bool in_plt, DoneList *dlp)
{
    const Elf_Sym *def, *def_w;
    const Needed_Entry *n;
    const Obj_Entry *obj, *defobj, *defobj1;
    
    def = def_w = NULL;
    defobj = NULL;
    for (n = needed; n != NULL; n = n->next) {
        if ((obj = n->obj) == NULL ||
            donelist_check(dlp, obj) ||
            (def = Tcl_SymlookObj(name, hash, obj, in_plt)) == NULL)
                continue;
        defobj = obj;
        if (ELF_ST_BIND(def->st_info) != STB_WEAK) {
            *defobj_out = defobj;
            return (def);
	}
    }
    /*
     * There we come when either symbol definition is not found in
     * directly needed objects, or found symbol is weak.
     */
    for (n = needed; n != NULL; n = n->next) {
        if ((obj = n->obj) == NULL)
            continue;
        def_w = Tcl_SymlookNeeded(name, hash, obj->needed, &defobj1,
			       in_plt, dlp);
        if (def_w == NULL)
            continue;
        if (def == NULL || ELF_ST_BIND(def_w->st_info) != STB_WEAK) {
            def = def_w;
            defobj = defobj1;
        }
        if (ELF_ST_BIND(def_w->st_info) != STB_WEAK)
            break;
    }
    if (def != NULL)
        *defobj_out = defobj;
    return def;
}

/*
 * Search the symbol table of a single shared object for a symbol of
 * the given name.  Returns a pointer to the symbol, or NULL if no
 * definition was found.
 *
 * The symbol's hash value is passed in for efficiency reasons; that
 * eliminates many recomputations of the hash value.
 */
const Elf_Sym *
Tcl_SymlookObj(const char *name, unsigned long hash, const Obj_Entry *obj,
  bool in_plt)
{
    if (obj->buckets != NULL) {
	unsigned long symnum = obj->buckets[hash % obj->nbuckets];

	while (symnum != STN_UNDEF) {
	    const Elf_Sym *symp;
	    const char *strp;

	    if (symnum >= obj->nchains)
		return NULL;	/* Bad object */
	    symp = obj->symtab + symnum;
	    strp = obj->strtab + symp->st_name;

	    if (name[0] == strp[0] && strcmp(name, strp) == 0) {
		return symp->st_shndx != SHN_UNDEF ||
		  (!in_plt && symp->st_value != 0 &&
		  ELF_ST_TYPE(symp->st_info) == STT_FUNC) ? symp : NULL;
	    }
	    symnum = obj->chains[symnum];
	}
    }
    return NULL;
}

/*
 * Unload a dlopened object and its dependencies from memory and from
 * our data structures.  It is assumed that the DAG rooted in the
 * object has already been unreferenced, and that the object has a
 * reference count of 0.
 */
static void
Tcl_UnloadObject(Obj_Entry *root)
{
    Obj_Entry *obj, *obj_tmp;
    Obj_Entry **linkp;

    assert(root->refcount == 0);

    /*
     * Pass over the DAG removing unreferenced objects from
     * appropriate lists.
     */ 
    Tcl_UnlinkObject(root);

    /* Unmap all objects that are no longer referenced. */
    obj = obj_list->next;
    while (obj != NULL) {
	if (obj->refcount == 0) {
	    printf("unloading \"%s\" %u\n", obj->path, obj);
	    munmap(obj->mapbase, obj->mapsize);
	    obj->prev->next = obj->next;
	    if (!obj->next) {
		obj_tail = obj->prev;
	    } else {
		obj->next->prev = obj->prev;
	    }
	    obj_count--;
	    obj_tmp = obj;
	    obj = obj->next;
	    Tcl_ObjFree(obj_tmp);
	} else
	    obj = obj->next;
    }
}

static void
Tcl_UnlinkObject(Obj_Entry *root)
{
    const Needed_Entry *needed;
    Objlist_Entry *elm;

    if (root->refcount == 0) {
	/* Remove the object from the RTLD_GLOBAL list. */
	objlist_remove(&list_global, root);

    	/* Remove the object from all objects' DAG lists. */
    	STAILQ_FOREACH(elm, &root->dagmembers , link)
	    objlist_remove(&elm->obj->dldags, root);
    }

    for (needed = root->needed;  needed != NULL;  needed = needed->next)
	if (needed->obj != NULL)
	    Tcl_UnlinkObject(needed->obj);
}

static void
Tcl_UnrefDag(Obj_Entry *root)
{
    const Needed_Entry *needed;

    if (root->refcount == 0)
	return;
    root->refcount--;
    if (root->refcount == 0)
	for (needed = root->needed;  needed != NULL;  needed = needed->next)
	    if (needed->obj != NULL)
		Tcl_UnrefDag(needed->obj);
}




/*
 * Allocate TLS block for module with given index.
 */


static Elf_Ehdr *Tcl_GetElfHeader (int, const char *);
static int convert_prot(int);	/* Elf flags -> mmap protection */

/**
 * Map a shared object into memory. The shared object is located in a file on disk
 * \param fd  file descriptor, which must be open on the object and positioned at its beginning.
 * \param path a pathname that is used only for error messages.
 * \return The return value is a pointer to a newly-allocated Obj_Entry structure
 * for the shared object.  Returns NULL on failure.
 */
Obj_Entry *
Tcl_MapObjectFromFile(int fd, const char *path, const struct stat *sb)
{
    Obj_Entry *obj;
    Elf_Ehdr *hdr;
    int i;
    Elf_Phdr *phdr;
    Elf_Phdr *phlimit;
    Elf_Phdr **segs;
    int nsegs;
    Elf_Phdr *phdyn;
    Elf_Phdr *phphdr;
    Elf_Phdr *phinterp;
    Elf_Phdr *phtls;
    caddr_t mapbase;
    size_t mapsize;
    Elf_Off base_offset;
    Elf_Addr base_vaddr;
    Elf_Addr base_vlimit;
    caddr_t base_addr;
    Elf_Off data_offset;
    Elf_Addr data_vaddr;
    Elf_Addr data_vlimit;
    caddr_t data_addr;
    int data_prot;
    int data_flags;
    Elf_Addr clear_vaddr;
    caddr_t clear_addr;
    caddr_t clear_page;
    size_t nclear;
    Elf_Addr bss_vaddr;
    Elf_Addr bss_vlimit;
    caddr_t bss_addr;

    hdr = Tcl_GetElfHeader(fd, path);
    if (hdr == NULL)
	return NULL;

    /*
     * Scan the program header entries, and save key information.
     *
     * We rely on there being exactly two load segments, text and data,
     * in that order.
     */
    phdr = (Elf_Phdr *) ((char *)hdr + hdr->e_phoff);
    phlimit = phdr + hdr->e_phnum;
    nsegs = -1;
    phdyn = phphdr = phinterp = phtls = NULL;
    segs = alloca(sizeof(segs[0]) * hdr->e_phnum);
    while (phdr < phlimit) {
	switch (phdr->p_type) {

	case PT_INTERP:
	    phinterp = phdr;
	    break;

	case PT_LOAD:
	    segs[++nsegs] = phdr;
    	    if (segs[nsegs]->p_align < PAGE_SIZE) {
		printf("%s: PT_LOAD segment %d not page-aligned",
		    path, nsegs);
		return NULL;
	    }
	    break;

	case PT_PHDR:
	    phphdr = phdr;
	    break;

	case PT_DYNAMIC:
	    phdyn = phdr;
	    break;
	case PT_TLS:
	    phtls = phdr;
	    break;
	}

	++phdr;
    }
    if (phdyn == NULL) {
	printf("%s: object is not dynamically-linked", path);
	return NULL;
    }

    if (nsegs < 0) {
	printf("%s: too few PT_LOAD segments", path);
	return NULL;
    }

    /*
     * Map the entire address space of the object, to stake out our
     * contiguous region, and to establish the base address for relocation.
     */
    base_offset = trunc_page(segs[0]->p_offset);
    base_vaddr = trunc_page(segs[0]->p_vaddr);
    base_vlimit = round_page(segs[nsegs]->p_vaddr + segs[nsegs]->p_memsz);
    mapsize = base_vlimit - base_vaddr;
    base_addr = hdr->e_type == ET_EXEC ? (caddr_t) base_vaddr : NULL;

    mapbase = mmap(base_addr, mapsize, convert_prot(segs[0]->p_flags),
      MAP_PRIVATE, fd, base_offset);
    if (mapbase == (caddr_t) -1) {
	printf("%s: mmap of entire address space failed: %s",
	  path, strerror(errno));
	return NULL;
    }
    if (base_addr != NULL && mapbase != base_addr) {
	printf("%s: mmap returned wrong address: wanted %p, got %p",
	  path, base_addr, mapbase);
	munmap(mapbase, mapsize);
	return NULL;
    }

    for (i = 0; i <=  nsegs; i++) {
	/* Overlay the segment onto the proper region. */
	data_offset = trunc_page(segs[i]->p_offset);
	data_vaddr = trunc_page(segs[i]->p_vaddr);
	data_vlimit = round_page(segs[i]->p_vaddr + segs[i]->p_filesz);
	data_addr = mapbase + (data_vaddr - base_vaddr);
	data_prot = convert_prot(segs[i]->p_flags);
	data_flags = MAP_PRIVATE | MAP_FIXED;
	/* Do not call mmap on the first segment - this is redundant */
	if (i && mmap(data_addr, data_vlimit - data_vaddr, data_prot,
	  data_flags, fd, data_offset) == (caddr_t) -1) {
	    printf("%s: mmap of data failed: %s", path, strerror(errno));
	    return NULL;
	}

	/* Clear any BSS in the last page of the segment. */
	clear_vaddr = segs[i]->p_vaddr + segs[i]->p_filesz;
	clear_addr = mapbase + (clear_vaddr - base_vaddr);
	clear_page = mapbase + (trunc_page(clear_vaddr) - base_vaddr);
	if ((nclear = data_vlimit - clear_vaddr) > 0) {
	    /* Make sure the end of the segment is writable */
	    if ((data_prot & PROT_WRITE) == 0) {
		if (mprotect(clear_page, PAGE_SIZE, data_prot|PROT_WRITE) < 0) {
			printf("%s: mprotect failed: %s", path,
			    strerror(errno));
			return NULL;
		}
	    }

	    memset(clear_addr, 0, nclear);

	    /*
	     * reset the data protection back, enable the segment to be
	     * coredumped since we modified it.
	     */
	    if ((data_prot & PROT_WRITE) == 0) {
		madvise(clear_page, PAGE_SIZE, MADV_CORE);
		mprotect(clear_page, PAGE_SIZE, data_prot);
	    }
	}

	/* Overlay the BSS segment onto the proper region. */
	bss_vaddr = data_vlimit;
	bss_vlimit = round_page(segs[i]->p_vaddr + segs[i]->p_memsz);
	bss_addr = mapbase +  (bss_vaddr - base_vaddr);
	if (bss_vlimit > bss_vaddr) {	/* There is something to do */
	    if (mmap(bss_addr, bss_vlimit - bss_vaddr, data_prot,
		MAP_PRIVATE|MAP_FIXED|MAP_ANON, -1, 0) == (caddr_t) -1) {
		    printf("%s: mmap of bss failed: %s", path,
			strerror(errno));
		return NULL;
	    }
	}
    }

    obj = Tcl_ObjNew();
    if (sb != NULL) {
	obj->dev = sb->st_dev;
	obj->ino = sb->st_ino;
    }
    obj->mapbase = mapbase;
    obj->mapsize = mapsize;
    obj->textsize = round_page(segs[0]->p_vaddr + segs[0]->p_memsz) -
      base_vaddr;
    obj->vaddrbase = base_vaddr;
    obj->relocbase = mapbase - base_vaddr;
    obj->dynamic = (const Elf_Dyn *) (obj->relocbase + phdyn->p_vaddr);
    if (hdr->e_entry != 0)
	obj->entry = (caddr_t) (obj->relocbase + hdr->e_entry);
    if (phphdr != NULL) {
	obj->phdr = (const Elf_Phdr *) (obj->relocbase + phphdr->p_vaddr);
	obj->phsize = phphdr->p_memsz;
    }
    if (phinterp != NULL)
	obj->interp = (const char *) (obj->relocbase + phinterp->p_vaddr);
    if (phtls != NULL) {
	tls_dtv_generation++;
	obj->tlsindex = ++tls_max_index;
	obj->tlssize = phtls->p_memsz;
	obj->tlsalign = phtls->p_align;
	obj->tlsinitsize = phtls->p_filesz;
	obj->tlsinit = mapbase + phtls->p_vaddr;
    }
    return obj;
}

/**
 * Map a shared object into memory. The shared object is located in memory
 * \param buffer pointer in memory where the shared object is located
 * \param nbytes number of bytes of the shared object
 * \return The return value is a pointer to a newly-allocated Obj_Entry structure
 * for the shared object.  Returns NULL on failure.
 */
Obj_Entry* Tcl_MapObjectFromMemory(const void *buffer, int nbytes)
{
	Obj_Entry *obj;
	int i;
	Elf32_Ehdr *hdr;
	Elf32_Phdr *phdr;
	Elf32_Phdr *phlimit;
	Elf32_Phdr *phinterp;
	Elf32_Phdr *phphdr;
	Elf32_Phdr *phdyn;
	Elf32_Phdr *phtls;
	Elf32_Phdr **segs;
	int nsegs;
	caddr_t mapbase = 0;
	Elf32_Off base_offset;
	Elf32_Addr base_vaddr;
	Elf32_Addr base_vlimit;
	caddr_t base_addr = 0;
	Elf32_Off data_offset;
	Elf32_Addr data_vaddr;
	Elf32_Addr data_vlimit;
	caddr_t data_addr;
	Elf_Addr clear_vaddr;
	caddr_t clear_addr;
	caddr_t clear_page;
	Elf_Addr bss_vaddr;
	Elf_Addr bss_vlimit;
	caddr_t bss_addr;
	int data_prot;
	int data_flags;
	size_t mapsize;
	size_t nclear;

	
	/* First we have to check if the memory image in buffer really represents an ELF file */
	if (nbytes < sizeof(Elf32_Ehdr)) {
		printf("Invalid file format\n");
		return NULL;
	}

	hdr = (Elf32_Ehdr *)buffer;
	if (memcmp(hdr, ELFMAG, SELFMAG) != 0) {
		printf("Invalid file format\n");
		return NULL;
	}
	if (hdr->e_ident[EI_CLASS] != ELFCLASS32 ||
	    hdr->e_ident[EI_DATA] != ELFDATA2LSB) {
		printf("Unsupported file layout\n");
		return NULL;
	}
	if (hdr->e_ident[EI_VERSION] != EV_CURRENT) {
		printf("Unsupported file version\n");
		return NULL;
	}
	if (hdr->e_type != ET_EXEC && hdr->e_type != ET_DYN) {
		printf("Unsupported file type\n");
		return NULL;
	}
	if (hdr->e_machine != EM_386) {
		printf("Unsupported machine\n");
		return NULL;
	}
	if (hdr->e_phentsize != sizeof(Elf32_Phdr)) {
		printf("Invalid shared object: e_phentsize != sizeof(Elf_Phdr)\n");
		return NULL;
	}
	if (hdr->e_phoff + hdr->e_phnum * sizeof(Elf32_Phdr) > (size_t)nbytes) {
		printf("Program header too large\n");
		return NULL;
	}

	phdr = (Elf32_Phdr *) ((char *)hdr + hdr->e_phoff);
	phlimit = phdr + hdr->e_phnum;
	 phdyn = phphdr = phinterp = phtls = NULL;
	segs = alloca(sizeof(segs[0]) * hdr->e_phnum);
	nsegs = -1;
	while (phdr < phlimit) {
		switch (phdr->p_type) {

		  case PT_INTERP:
			phinterp = phdr;
			break;

		  case PT_LOAD:
			segs[++nsegs] = phdr;
			if (segs[nsegs]->p_align < 4096) {
				printf("PT_LOAD segment %d not page-aligned\n", nsegs);
				return NULL;
			}
			break;

		  case PT_PHDR:
			phphdr = phdr;
			break;

		  case PT_DYNAMIC:
			phdyn = phdr;
			break;

		  case PT_TLS:
			phtls = phdr;
			break;

		}
		++phdr;
	}

	if (phdyn == NULL) {
		printf("Object is not dynamically-linked\n");
		return NULL;
	}

	if (nsegs < 0) {
		printf("Too few PT_LOAD segments");
		return NULL;
	}

	base_offset = trunc_page(segs[0]->p_offset);
	base_vaddr = trunc_page(segs[0]->p_vaddr);
	base_vlimit = round_page(segs[nsegs]->p_vaddr + segs[nsegs]->p_memsz);
	mapsize = base_vlimit - base_vaddr;
	base_addr = hdr->e_type == ET_EXEC ? (caddr_t) base_vaddr : NULL;
	mapbase = mmap(base_addr, mapsize, PROT_READ|PROT_EXEC|PROT_WRITE,
          MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);

	/*if (!(mapbase = malloc(mapsize))) {
		printf("Couldn't allocate memory\n");
		return NULL;
	}*/
	
	memcpy(mapbase, buffer + base_offset, mapsize);

	for(i = 0; i <= nsegs; ++i) {
		int res;
		data_offset = trunc_page(segs[i]->p_offset);
		data_vaddr = trunc_page(segs[i]->p_vaddr);
		data_vlimit = round_page(segs[i]->p_vaddr + segs[i]->p_filesz);
		data_addr = mapbase + (data_vaddr - base_vaddr);
		data_prot = convert_prot(segs[i]->p_flags);
		data_flags = MAP_PRIVATE | MAP_FIXED | MAP_ANONYMOUS;
		
		if (i && mmap(data_addr, data_vlimit - data_vaddr, data_prot,
		  data_flags, -1, data_offset) == (caddr_t) -1) {
			printf("mmap of data failed: %s", strerror(errno));
			return NULL;
		}
		memcpy(data_addr, buffer + data_offset, data_vlimit - data_vaddr);
		/* Clear any BSS in the last page of the segment. */
		clear_vaddr = segs[i]->p_vaddr + segs[i]->p_filesz;
		clear_addr = mapbase + (clear_vaddr - base_vaddr);
		clear_page = mapbase + (trunc_page(clear_vaddr) - base_vaddr);

		if ((nclear = data_vlimit - clear_vaddr) > 0) {

			/* Make sure the end of the segment is writable */
			if ((data_prot & PROT_WRITE) == 0) {
				if (mprotect(clear_page, PAGE_SIZE, data_prot|PROT_WRITE) < 0) {
					return NULL;
				}
			}
		
			memset(clear_addr, 0, nclear);
		
			/*
			* reset the data protection back, enable the segment to be
			* coredumped since we modified it.
			*/
			if ((data_prot & PROT_WRITE) == 0) {
				madvise(clear_page, PAGE_SIZE, MADV_CORE);
				mprotect(clear_page, PAGE_SIZE, data_prot);
			}
		}

		/* Overlay the BSS segment onto the proper region. */
		bss_vaddr = data_vlimit;
		bss_vlimit = round_page(segs[i]->p_vaddr + segs[i]->p_memsz);
		bss_addr = mapbase +  (bss_vaddr - base_vaddr);
		if (bss_vlimit > bss_vaddr) {	/* There is something to do */
			if (mmap(bss_addr, bss_vlimit - bss_vaddr, data_prot,
				MAP_PRIVATE|MAP_FIXED|MAP_ANON, -1, 0) == (caddr_t) -1) {
				return NULL;
			}
		}
	}

	obj = Tcl_ObjNew();
	obj->mapbase = mapbase;
	obj->mapsize = mapsize;
	obj->textsize = round_page(segs[0]->p_vaddr + segs[0]->p_memsz) -
	base_vaddr;
	obj->vaddrbase = base_vaddr;
	obj->relocbase = mapbase - base_vaddr;
	obj->dynamic = (const Elf_Dyn *) (obj->relocbase + phdyn->p_vaddr);

	if (hdr->e_entry != 0)
		obj->entry = (caddr_t) (obj->relocbase + hdr->e_entry);
	if (phphdr != NULL) {
		obj->phdr = (const Elf_Phdr *) (obj->relocbase + phphdr->p_vaddr);
		obj->phsize = phphdr->p_memsz;
	}

	if (phinterp != NULL)
		obj->interp = (const char *) (obj->relocbase + phinterp->p_vaddr);

	if (phtls != NULL) {
		tls_dtv_generation++;
		obj->tlsindex = ++tls_max_index;
		obj->tlssize = phtls->p_memsz;
		obj->tlsalign = phtls->p_align;
		obj->tlsinitsize = phtls->p_filesz;
		obj->tlsinit = mapbase + phtls->p_vaddr;
	}
	return obj;

}

/**
 * One of the sections in ELF file is 'section header'. This function reads this section
 * from a given file and returns Elf_Ehdr struct.
 * \param fd descriptor of ELF file
 * \param path path to this file (it is used for error reports)
 * \return Elf_Ehdr struct filled with proper data from the file, or NULL if something went wrong
 */
static Elf_Ehdr *
Tcl_GetElfHeader (int fd, const char *path)
{
    static union {
	Elf_Ehdr hdr;
	char buf[PAGE_SIZE];
    } u;
    ssize_t nbytes;

    if ((nbytes = read(fd, u.buf, PAGE_SIZE)) == -1) {
	printf("%s: read error: %s", path, strerror(errno));
	return NULL;
    }

    /* Make sure the file is valid */
    if (nbytes < (ssize_t)sizeof(Elf_Ehdr)) {
	printf("%s: invalid file format", path);
	return NULL;
    }
    if (memcmp(&u.hdr, ELFMAG, SELFMAG) != 0) {
	printf("Invalid file format\n");
	return NULL;
    }
    if (u.hdr.e_ident[EI_CLASS] != ELF_TARG_CLASS
      || u.hdr.e_ident[EI_DATA] != ELF_TARG_DATA) {
	printf("%s: unsupported file layout", path);
	return NULL;
    }
    if (u.hdr.e_ident[EI_VERSION] != EV_CURRENT
      || u.hdr.e_version != EV_CURRENT) {
	printf("%s: unsupported file version", path);
	return NULL;
    }
    if (u.hdr.e_type != ET_EXEC && u.hdr.e_type != ET_DYN) {
	printf("%s: unsupported file type", path);
	return NULL;
    }
    if (u.hdr.e_machine != ELF_TARG_MACH) {
	printf("%s: unsupported machine", path);
	return NULL;
    }

    /*
     * We rely on the program header being in the first page.  This is
     * not strictly required by the ABI specification, but it seems to
     * always true in practice.  And, it simplifies things considerably.
     */
    if (u.hdr.e_phentsize != sizeof(Elf_Phdr)) {
	printf(
	  "%s: invalid shared object: e_phentsize != sizeof(Elf_Phdr)", path);
	return NULL;
    }
    if (u.hdr.e_phoff + u.hdr.e_phnum * sizeof(Elf_Phdr) > (size_t)nbytes) {
	printf("%s: program header too large", path);
	return NULL;
    }

    return (&u.hdr);
}

/**
 * Free memory of a shared object structure. It also frees memory of structures
 * that depended on the shared object.
 * \param obj the shared object that will be freed
 */
void
Tcl_ObjFree(Obj_Entry *obj)
{
    Objlist_Entry *elm;

    free(obj->origin_path);
    free(obj->path);
    while (obj->needed != NULL) {
	Needed_Entry *needed = obj->needed;
	obj->needed = needed->next;
	free(needed);
    }
    while (!STAILQ_EMPTY(&obj->dldags)) {
	elm = STAILQ_FIRST(&obj->dldags);
	STAILQ_REMOVE_HEAD(&obj->dldags, link);
	free(elm);
    }
    while (!STAILQ_EMPTY(&obj->dagmembers)) {
	elm = STAILQ_FIRST(&obj->dagmembers);
	STAILQ_REMOVE_HEAD(&obj->dagmembers, link);
	free(elm);
    }
    free(obj);
}

/**
 * Create a new shared object and initialize some of its structures
 * \return returns newly allocated Obj_Entry structure
 */
Obj_Entry *
Tcl_ObjNew(void)
{
    Obj_Entry *obj;

    obj = CNEW(Obj_Entry);
    STAILQ_INIT(&obj->dldags);
    STAILQ_INIT(&obj->dagmembers);
    return obj;
}

/**
 * Given a set of ELF protection flags, return the corresponding protection
 * flags for MMAP.
 */
static int
convert_prot(int elfflags)
{
    int prot = 0;
    if (elfflags & PF_R) {
	prot |= PROT_READ;
	printf("PROT READ\n");
    }
    if (elfflags & PF_W) {
	prot |= PROT_WRITE;
	printf("PROT_WRITE\n");
    }
    if (elfflags & PF_X) {
	prot |= PROT_EXEC;
	printf("PROT_EXEC\n");
    }
    return prot;
}

/** Initialize the special GOT entries. */
void
Tcl_InitPltGot_i386(Obj_Entry *obj)
{
    if (obj->pltgot != NULL) {
	obj->pltgot[1] = (Elf_Addr) obj;
    }
}

/** Process the non-PLT relocations. */
int
Tcl_RelocNonPlt_i386(Obj_Entry *obj, Obj_Entry *obj_rtld)
{
	const Elf_Rel *rellim;
	const Elf_Rel *rel;
	SymCache *cache;
	int bytes;
	int r = -1;

	bytes = obj->nchains * sizeof(SymCache);
	/*
	 * The dynamic loader may be called from a thread, we have
	 * limited amounts of stack available so we cannot use alloca().
	 */
	cache = mmap(NULL, bytes, PROT_READ|PROT_WRITE, MAP_ANON, -1, 0);
	if (cache == MAP_FAILED)
	    cache = NULL;
	rellim = (const Elf_Rel *) ((caddr_t) obj->rel + obj->relsize);

	for (rel = obj->rel;  rel < rellim;  rel++) {
	    Elf_Addr *where = (Elf_Addr *) (obj->relocbase + rel->r_offset);

	    switch (ELF_R_TYPE(rel->r_info)) {

	    case R_386_NONE:
		break;

	    case R_386_32:
		{
		    const Elf_Sym *def;
		    const Obj_Entry *defobj;

		    def = Tcl_FindSymdef(ELF_R_SYM(rel->r_info), obj, &defobj,
		      false, cache);
		    if (def == NULL)
			goto done;

		    *where += (Elf_Addr) (defobj->relocbase + def->st_value);
		}
		break;

	    case R_386_PC32:
		/*
		 * I don't think the dynamic linker should ever see this
		 * type of relocation.  But the binutils-2.6 tools sometimes
		 * generate it.
		 */
		{
		    const Elf_Sym *def;
		    const Obj_Entry *defobj;

		    def = Tcl_FindSymdef(ELF_R_SYM(rel->r_info), obj, &defobj,
		      false, cache);
		    if (def == NULL)
			goto done;

		    *where +=
		      (Elf_Addr) (defobj->relocbase + def->st_value) -
		      (Elf_Addr) where;
		}
		break;

	    case R_386_COPY:
		/*
		 * These are deferred until all other relocations have
		 * been done.  All we do here is make sure that the COPY
		 * relocation is not in a shared library.  They are allowed
		 * only in executable files.
		 */

		if (!obj->mainprog) {
		    printf("%s: Unexpected R_386_COPY relocation"
		      " in shared library", obj->path);
		    goto done;
		}
		break;

	    case R_386_GLOB_DAT:
		{
		    const Elf_Sym *def;
		    const Obj_Entry *defobj;

		    def = Tcl_FindSymdef(ELF_R_SYM(rel->r_info), obj, &defobj,
		      false, cache);
		    if (def == NULL)
			goto done;
		    *where = (Elf_Addr) (defobj->relocbase + def->st_value);

		}
		break;

	    case R_386_RELATIVE:

		*where += (Elf_Addr) obj->relocbase;
		break;

	    case R_386_TLS_DTPMOD32:
		{
		    const Elf_Sym *def;
		    const Obj_Entry *defobj;

		    def = Tcl_FindSymdef(ELF_R_SYM(rel->r_info), obj, &defobj,
		      false, cache);
		    if (def == NULL)
			goto done;

		    *where += (Elf_Addr) defobj->tlsindex;
		}
		break;

	    case R_386_TLS_DTPOFF32:
		{
		    const Elf_Sym *def;
		    const Obj_Entry *defobj;

		    def = Tcl_FindSymdef(ELF_R_SYM(rel->r_info), obj, &defobj,
		      false, cache);
		    if (def == NULL)
			goto done;

		    *where += (Elf_Addr) def->st_value;
		}
		break;

	    default:
		printf("%s: Unsupported relocation type %d"
		  " in non-PLT relocations\n", obj->path,
		  ELF_R_TYPE(rel->r_info));
		goto done;
	    }
	}
	r = 0;
done:
	if (cache)
	    munmap(cache, bytes);
	return(r);
}

/** Process the PLT relocations. */
int
Tcl_RelocPlt_i386(Obj_Entry *obj)
{
    const Elf_Rel *rellim;
    const Elf_Rel *rel;

    rellim = (const Elf_Rel *)((const char *)obj->pltrel + obj->pltrelsize);
    for (rel = obj->pltrel;  rel < rellim;  rel++) {
	Elf_Addr *where;

	assert(ELF_R_TYPE(rel->r_info) == R_386_JMP_SLOT);

	/* Relocate the GOT slot pointing into the PLT. */
	where = (Elf_Addr *)(obj->relocbase + rel->r_offset);
	*where += (Elf_Addr)obj->relocbase;
    }
    return 0;
}

/** Relocate the jump slots in an object. */
int
Tcl_RelocJmpslots_i386(Obj_Entry *obj)
{
    const Elf_Rel *rellim;
    const Elf_Rel *rel;

    if (obj->jmpslots_done)
	return 0;
    rellim = (const Elf_Rel *)((const char *)obj->pltrel + obj->pltrelsize);
    for (rel = obj->pltrel;  rel < rellim;  rel++) {
	Elf_Addr *where;
	const Elf_Sym *def;
	const Obj_Entry *defobj;

	assert(ELF_R_TYPE(rel->r_info) == R_386_JMP_SLOT);
	where = (Elf_Addr *)(obj->relocbase + rel->r_offset);
	def = Tcl_FindSymdef(ELF_R_SYM(rel->r_info), obj, &defobj, true, NULL);
	if (def == NULL)
	    return -1;
	reloc_jmpslot(where, (Elf_Addr)(defobj->relocbase + def->st_value));
    }
    obj->jmpslots_done = true;
    return 0;
}

void *
xcalloc(size_t size)
{
    return memset(xmalloc(size), 0, size);
}

void *
xmalloc(size_t size)
{
    void *p = malloc(size);
    if (p == NULL)
	err(1, "Out of memory");
    return p;
}

void *
xrealloc(void *cp, size_t nbytes)
{
    void *p = realloc(cp, nbytes);
    if (p == NULL)
	err(1, "Out of memory");
    return p;
}

char *
xstrdup(const char *s)
{
    char *p = strdup(s);
    if (p == NULL)
	err(1, "Out of memory");
    return p;
}

int main(int argc, char** argv)
{
	int fd;
	char *addr;
	void *fun, *fun2;
	void *handle;
	struct stat stat;
	int (*greeting)();
	int (*retfun)();
	int (*argfun)(int);

	obj_main = Tcl_ObjNew();
	obj_list = obj_main;
	obj_tail = obj_main;

	fd = open(argv[1], O_RDWR);
	fstat(fd, &stat);

	addr = mmap(0, stat.st_size, PROT_READ | PROT_WRITE, MAP_PRIVATE, fd, 0);

	fun = Tcl_Dlopen(addr, stat.st_size, RTLD_NOW);

	munmap(addr, stat.st_size);
	close(fd);

	greeting = Tcl_Dlsym(fun, "greet");
	
	if(greeting) {
		printf("Exectute greeting\n");
		(*greeting)();
	}

	if (argc > 2) {
		fd = open(argv[2], O_RDWR);
		printf("FD = %d\n", fd);
		fstat(fd, &stat);
		addr = mmap(0, stat.st_size, PROT_READ | PROT_WRITE, MAP_PRIVATE, fd, 0);
		fun2 = Tcl_Dlopen(addr, stat.st_size, RTLD_NOW);
		munmap(addr, stat.st_size);
		close(fd);
		argfun = Tcl_Dlsym(fun2, "greet");
		if(argfun) {
			printf("Exectute greeting %d\n", fd);
			fd = (*argfun)(fd);
		}
	}

	Tcl_Dlclose(fun);
	if (argc > 2)
		Tcl_Dlclose(fun2);
	assert(obj_tail == obj_list);
	return 0;
}

