/*
    tcl-fuse - Tcl FUSE Interface
    A Tcl interface to the Linux kernel's FUSE subsystem.

    Copyright (C) 2008  Alexandros Stergiakis

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * convert.c
 *
 */

#include <stdio.h> // snprintf
#include <stdlib.h> // atoll
#include <string.h> // memcpy
#include <sys/types.h> // stat types
#include <sys/stat.h> // stat types
#include <unistd.h> // stat types
#include <errno.h> // error codes
#include <fcntl.h> // open flags

#include <tcl.h>
#include <fuse.h>

#include "tclfuse.h"

#define BUFSIZE 1000

/* Returns NULL if not recognized */
int convert__errno(Tcl_Interp *interp) {
    static const char *errcodes[] = {
        "E2BIG", "EACCES", "EADDRINUSE", "EADDRNOTAVAIL", "EAFNOSUPPORT", "EAGAIN",
        "EALREADY", "EBADF", "EBADMSG", "EBUSY", "ECANCELED", "ECHILD", "ECONNABORTED",
        "ECONNREFUSED", "ECONNRESET", "EDEADLK", "EDESTADDRREQ", "EDOM", "EDQUOT",
        "EEXIST", "EFAULT", "EFBIG", "EHOSTUNREACH", "EIDRM", "EILSEQ", "EINPROGRESS",
        "EINTR", "EINVAL", "EIO", "EISCONN", "EISDIR", "ELOOP", "EMFILE", "EMLINK",
        "EMSGSIZE", "EMULTIHOP", "ENAMETOOLONG", "ENETDOWN", "ENETRESET", "ENETUNREACH",
        "ENFILE", "ENOBUFS", "ENODATA", "ENODEV", "ENOENT", "ENOEXEC", "ENOLCK", "ENOLINK",
        "ENOMEM", "ENOMSG", "ENOPROTOOPT", "ENOSPC", "ENOSR", "ENOSTR", "ENOSYS", "ENOTCONN",
        "ENOTDIR", "ENOTEMPTY", "ENOTSOCK", "ENOTSUP", "ENOTTY", "ENXIO", "EOPNOTSUPP",
        "EOVERFLOW", "EPERM", "EPIPE", "EPROTO", "EPROTONOSUPPORT", "EPROTOTYPE", "ERANGE",
        "EROFS", "ESPIPE", "ESRCH", "ESTALE", "ETIME", "ETIMEDOUT", "ETXTBSY", "EWOULDBLOCK",
        "EXDEV", NULL
    };

    Tcl_Obj *options = Tcl_GetReturnOptions(interp, TCL_ERROR);
    Tcl_Obj *key = Tcl_NewStringObj("-errorcode", -1);
    Tcl_Obj *list;
    Tcl_Obj *obj;
    Tcl_IncrRefCount(key);
    Tcl_DictObjGet(NULL, options, key, &list);
    Tcl_DecrRefCount(key);
    Tcl_ListObjIndex(NULL, list, 0, &obj);
    if (strcmp("POSIX", Tcl_GetString(obj)) == 0)
        Tcl_ListObjIndex(NULL, list, 1, &obj);
    else
        return 0;

    int index;
    if (Tcl_GetIndexFromObj(NULL, obj, errcodes, "", TCL_EXACT, &index) != TCL_OK)
        return 0;

    switch (index) {
        case 0: return -E2BIG;
        case 1: return -EACCES;
        case 2: return -EADDRINUSE;
        case 3: return -EADDRNOTAVAIL;
        case 4: return -EAFNOSUPPORT;
        case 5: return -EAGAIN;
        case 6: return -EALREADY;
        case 7: return -EBADF;
        case 8: return -EBADMSG;
        case 9: return -EBUSY;
        case 10: return -ECANCELED;
        case 11: return -ECHILD;
        case 12: return -ECONNABORTED;
        case 13: return -ECONNREFUSED;
        case 14: return -ECONNRESET;
        case 15: return -EDEADLK;
        case 16: return -EDESTADDRREQ;
        case 17: return -EDOM;
        case 18: return -EDQUOT;
        case 19: return -EEXIST;
        case 20: return -EFAULT;
        case 21: return -EFBIG;
        case 22: return -EHOSTUNREACH;
        case 23: return -EIDRM;
        case 24: return -EILSEQ;
        case 25: return -EINPROGRESS;
        case 26: return -EINTR;
        case 27: return -EINVAL;
        case 28: return -EIO;
        case 29: return -EISCONN;
        case 30: return -EISDIR;
        case 31: return -ELOOP;
        case 32: return -EMFILE;
        case 33: return -EMLINK;
        case 34: return -EMSGSIZE;
        case 35: return -EMULTIHOP;
        case 36: return -ENAMETOOLONG;
        case 37: return -ENETDOWN;
        case 38: return -ENETRESET;
        case 39: return -ENETUNREACH;
        case 40: return -ENFILE;
        case 41: return -ENOBUFS;
        case 42: return -ENODATA;
        case 43: return -ENODEV;
        case 44: return -ENOENT;
        case 45: return -ENOEXEC;
        case 46: return -ENOLCK;
        case 47: return -ENOLINK;
        case 48: return -ENOMEM;
        case 49: return -ENOMSG;
        case 50: return -ENOPROTOOPT;
        case 51: return -ENOSPC;
        case 52: return -ENOSR;
        case 53: return -ENOSTR;
        case 54: return -ENOSYS;
        case 55: return -ENOTCONN;
        case 56: return -ENOTDIR;
        case 57: return -ENOTEMPTY;
        case 58: return -ENOTSOCK;
        case 59: return -ENOTSUP;
        case 60: return -ENOTTY;
        case 61: return -ENXIO;
        case 62: return -EOPNOTSUPP;
        case 63: return -EOVERFLOW;
        case 64: return -EPERM;
        case 65: return -EPIPE;
        case 66: return -EPROTO;
        case 67: return -EPROTONOSUPPORT;
        case 68: return -EPROTOTYPE;
        case 69: return -ERANGE;
        case 70: return -EROFS;
        case 71: return -ESPIPE;
        case 72: return -ESRCH;
        case 73: return -ESTALE;
        case 74: return -ETIME;
        case 75: return -ETIMEDOUT;
        case 76: return -ETXTBSY;
        case 77: return -EWOULDBLOCK;
        case 78: return -EXDEV;
    }
    return 0;
}

/* Returns NULL on failure */
Tcl_Obj *convert__fuse_file_info(struct fuse_file_info *fi) {
    char buffer[BUFSIZE];
    // Note: The items on the second line on each of the following arrays do not have an equivalent on the other array.
    static char *flags_str[] = {"RDONLY", "WRONLY", "RDWR", "APPEND", "CREAT", "EXCL", "NOCTTY", "NONBLOCK", "TRUNC"}; 
                               // "BINARY"
    static int flags_int[]   = {O_RDONLY, O_WRONLY, O_RDWR, O_APPEND, O_CREAT, O_EXCL, O_NOCTTY, O_NONBLOCK, O_TRUNC};
                                // O_ASYNC,  O_DIRECT, O_DIRECTORY, O_LARGEFILE, O_NOATIME, O_SYNC

    snprintf(buffer, BUFSIZE, "%s %s %ld %s %ld %s %ld %s %ld %s %ld %s %ld", 
        "flags {}",
        "writepage", (long) fi->writepage,
        "direct_io", (long) fi->direct_io,
        "keep_cache", (long) fi->keep_cache,
        "flush", (long) fi->flush,
        "file_handle", (long) fi->fh,
        "lock_owner", (long) fi->lock_owner);

    Tcl_Obj *listPtr = Tcl_NewListObj(0, NULL);
    int i;
    for (i = 0; i < 9; i++) {
        if ((fi->flags & flags_int[i]) == flags_int[i]) {
            Tcl_Obj *objPtr = Tcl_NewStringObj(flags_str[i], -1);
            if (Tcl_ListObjAppendElement(NULL, listPtr, objPtr) == TCL_ERROR)
                return NULL;
        }
    }

    Tcl_Obj *dictPtr = Tcl_NewStringObj(buffer, -1);
    Tcl_Obj *keyPtr = Tcl_NewStringObj("flags", -1);
    if (Tcl_DictObjPut(NULL, dictPtr, keyPtr, listPtr) == TCL_ERROR) {
        return NULL;
    }
    
    return dictPtr;
}

/* Returns NULL on failure */
Tcl_Obj *convert__fuse_conn_info(struct fuse_conn_info *ci) {
    char buffer[BUFSIZE];

    snprintf(buffer, BUFSIZE, "%s %ld %s %ld %s %ld %s %ld %s %ld", 
        "proto_major", (long) ci->proto_major,
        "proto_minor", (long) ci->proto_minor,
        "async_read", (long) ci->async_read,
        "max_write", (long) ci->max_write,
        "max_readahead", (long) ci->max_readahead);

    return Tcl_NewStringObj(buffer, -1);
}

/* Returns NULL on failure */
Tcl_Obj *convert__fuse_context(struct fuse_context *context) {
    char buffer[BUFSIZE];

    snprintf(buffer, BUFSIZE, "%s %ld %s %ld %s %ld",
        "uid", (long) context->uid,
        "gid", (long) context->gid,
        "pid", (long) context->pid);

    return Tcl_NewStringObj(buffer, -1);
}

/* Returns 0 on failure, 1 on success */
int rconvert__stat(Tcl_Interp *interp, struct stat *st, Tcl_Obj *dictPtr) {
    Tcl_Obj *keyPtr;
    Tcl_Obj *valuePtr;
    long val;
    static char *attr[] = {"dev", "ino", "mode", "nlink", "uid", 
                           "gid", "rdev", "size", "blksize", "blocks", 
                           "atime", "mtime", "ctime", "type"};

    int i;
    for (i = 0; i < 13; i++) {
        keyPtr = Tcl_NewStringObj(attr[i], -1);
        if (Tcl_DictObjGet(interp, dictPtr, keyPtr, &valuePtr) == TCL_ERROR) {
            goto error; // Cannot convert to dictionary.
        }
        if (valuePtr == NULL) {
            // defaults in case a key is not defined
            val = 0;
        } else {
            if (Tcl_GetLongFromObj(interp, valuePtr, &val) == TCL_ERROR) {
                Tcl_AppendResult(interp, "Cannot parse value of file attribute \"", attr[i], "\"", (char *) NULL);
                goto error;
            }
        }
        Tcl_DecrRefCount(keyPtr);

        switch (i) {
            // TODO: can overflow, we don't range check
            case 0: st->st_dev = (dev_t) val; break;
            case 1: st->st_ino = (ino_t) val; break;
            case 2: st->st_mode = (mode_t) val; break;
            case 3: st->st_nlink = (nlink_t) val; break;
            case 4: st->st_uid = (uid_t) val; break;
            case 5: st->st_gid = (gid_t) val; break;
            case 6: st->st_rdev = (dev_t) val; break;
            case 7: st->st_size = (off_t) val; break;
            case 8: st->st_blksize = (size_t) val; break;
            case 9: st->st_blocks = (blkcnt_t) val; break;
            case 10: st->st_atime = (time_t) val; break;
            case 11: st->st_mtime = (time_t) val; break;
            case 12: st->st_ctime = (time_t) val; break;
        }
    }

    /* 
     * We handle the attribute "type" specially:
     */

    int index;
    static char *types[] = {"file", "directory", "characterSpecial", 
                            "blockSpecial", "fifo", "link", "socket", NULL};

    keyPtr = Tcl_NewStringObj(attr[i], -1);
    Tcl_DictObjGet(interp, dictPtr, keyPtr, &valuePtr);
    if (valuePtr == NULL) {
        // default in case a key is not defined
        index = 0;
    } else {
        if (Tcl_GetIndexFromObj(interp, valuePtr, types, "file type", 0, &index) != TCL_OK)
            goto error;
    }

    Tcl_DecrRefCount(keyPtr);

    switch (index) {
        case 0: st->st_mode |= S_IFREG; break;
        case 1: st->st_mode |= S_IFDIR; break;
        case 2: st->st_mode |= S_IFCHR; break;
        case 3: st->st_mode |= S_IFBLK; break;
        case 4: st->st_mode |= S_IFIFO; break;
        case 5: st->st_mode |= S_IFLNK; break;
        case 6: st->st_mode |= S_IFSOCK; break;
    }

    return 1; // success

    error:
    Tcl_DecrRefCount(keyPtr);
    return 0;
}

/* returns NULL if not match */
Tcl_Obj *convert__type(mode_t mode) {
    static char *types_str[] = {"file", "directory", "characterSpecial", 
                            "blockSpecial", "fifo", "link", "socket"};
    static int types_int[] = {S_IFREG, S_IFDIR, S_IFCHR, S_IFBLK, S_IFIFO, S_IFLNK, S_IFSOCK};

    int i;
    for (i = 0; i < 7; i++)
        if ((mode & types_int[i]) == types_int[i])
            return Tcl_NewStringObj(types_str[i], -1);

    return NULL;
}

/* returns NULL in case of error, empty list if not match */
Tcl_Obj *convert__access_mask(int mask) {
    static char *masks_str[] = {"read", "write", "execute"};
    static int masks_int[] = {R_OK, W_OK, X_OK};

    if ((mask & F_OK) == F_OK)
        return Tcl_NewStringObj("exists", -1);

    Tcl_Obj *listPtr = Tcl_NewListObj(0, NULL);
    int i;
    for (i = 0; i < 3; i++)
        if ((mask & masks_int[i]) == masks_int[i]) {
            Tcl_Obj *objPtr = Tcl_NewStringObj(masks_str[i], -1);
            if (Tcl_ListObjAppendElement(NULL, listPtr, objPtr) == TCL_ERROR)
                return NULL;
        }

    return listPtr;
}

/* returns 0 on failure, 1 on success. On failure error message is left in interp */
int rconvert__statvfs(Tcl_Interp *interp, struct statvfs *st, Tcl_Obj *dictPtr) {

    Tcl_Obj *keyPtr;
    Tcl_Obj *valuePtr;
    long val;
    static char *attr[] = {"bsize", "frsize", "blocks", "bfree", "bavail", 
                           "files", "ffree", "favail", "fsid", "flag", "namemax"};

    int i;
    for (i = 0; i < 11; i++) {
        keyPtr = Tcl_NewStringObj(attr[i], -1);
        if (Tcl_DictObjGet(interp, dictPtr, keyPtr, &valuePtr) == TCL_ERROR) {
            goto error; // Cannot convert to dictionary.
        }
        if (valuePtr == NULL) {
            // defaults in case a key is not defined
            val = 0;
        } else {
            if (Tcl_GetLongFromObj(interp, valuePtr, &val) == TCL_ERROR) {
                Tcl_AppendResult(interp, "Cannot parse value of file attribute \"", attr[i], "\"", (char *) NULL);
                goto error;
            }
        }
        Tcl_DecrRefCount(keyPtr);

        switch (i) {
            // TODO: can overflow, we don't range check
            case 0: st->f_bsize = (unsigned long) val; break;
            case 1: st->f_frsize = (unsigned long) val; break;
            case 2: st->f_blocks = (fsfilcnt_t) val; break;
            case 3: st->f_bfree = (fsfilcnt_t) val; break;
            case 4: st->f_bavail = (fsfilcnt_t) val; break;
            case 5: st->f_files = (fsfilcnt_t) val; break;
            case 6: st->f_ffree = (fsfilcnt_t) val; break;
            case 7: st->f_favail = (off_t) val; break;
            case 8: st->f_fsid = (unsigned long) val; break;
            case 9: st->f_flag = (unsigned long) val; break;
            case 10: st->f_namemax = (unsigned long) val; break;
        }
    }

    return 1; // success

    error:
    Tcl_DecrRefCount(keyPtr);
    return 0;
}

Tcl_Obj *convert__flock(struct flock *lock) {
    char buffer[BUFSIZE];

    char *type;
    switch (lock->l_type) {
        case F_RDLCK: type = "readlock"; break;
        case F_WRLCK: type = "writelock"; break;
        case F_UNLCK: type = "unlock"; break;
    }

    char *origin;
    switch (lock->l_whence) {
        case SEEK_SET: origin = "start"; break;
        case SEEK_CUR: origin = "current"; break;
        case SEEK_END: origin = "end"; break;
    }

    snprintf(buffer, BUFSIZE, "%s %s %s %s %s %ld %s %ld", 
        "type", type,
        "origin", origin,
        "offset", (long int) lock->l_start,
        "length", (long int) lock->l_len);

    return Tcl_NewStringObj(buffer, -1);
}

/* If not match, NULL is returned */
Tcl_Obj *convert__lockcmd(int cmd) {
    switch (cmd) {
        case F_GETLK: return Tcl_NewStringObj("getlock", -1);
        case F_SETLK: return Tcl_NewStringObj("setlock", -1);
        case F_SETLKW: return Tcl_NewStringObj("setlockblock", -1);
    }
    return NULL;
}
